@@ -3,28 +3,36 @@ from __future__ import unicode_literals
 from __future__ import print_function
 
 import os
+import sys
 import traceback
 import logging
-import click
+from time import time
 
+import click
+import sqlparse
 from prompt_toolkit import CommandLineInterface, AbortAction, Exit
+from prompt_toolkit.document import Document
 from prompt_toolkit.layout import Layout
 from prompt_toolkit.layout.prompt import DefaultPrompt
 from prompt_toolkit.layout.menus import CompletionsMenu
 from prompt_toolkit.history import FileHistory
-from pygments.lexers.sql import SqlLexer
+from pygments.lexers.sql import PostgresLexer
 
 from .packages.tabulate import tabulate
 from .packages.expanded import expanded_table
 from .packages.pgspecial import (CASE_SENSITIVE_COMMANDS,
         NON_CASE_SENSITIVE_COMMANDS, is_expanded_output)
+import pgcli.packages.pgspecial as pgspecial
 from .pgcompleter import PGCompleter
 from .pgtoolbar import PGToolbar
-from .pgstyle import PGStyle
+from .pgstyle import style_factory
 from .pgexecute import PGExecute
 from .pgbuffer import PGBuffer
 from .config import write_default_config, load_config
 from .key_bindings import pgcli_bindings
+from .encodingutils import utf8tounicode
+from .__init__ import __version__
+
 
 try:
     from urlparse import urlparse
@@ -33,9 +41,9 @@ except ImportError:
 from getpass import getuser
 from psycopg2 import OperationalError
 
-from collections import deque, namedtuple
+from collections import namedtuple
 
-#Query tuples are used for maintaining history
+# Query tuples are used for maintaining history
 Query = namedtuple('Query', ['query', 'successful', 'mutating'])
 
 class PGCli(object):
@@ -54,14 +62,24 @@ class PGCli(object):
 
         # Load config.
         c = self.config = load_config('~/.pgclirc', default_config)
-        self.smart_completion = c.getboolean('main', 'smart_completion')
         self.multi_line = c.getboolean('main', 'multi_line')
+        self.vi_mode = c.getboolean('main', 'vi')
+        pgspecial.TIMING_ENABLED = c.getboolean('main', 'timing')
+        self.table_format = c.get('main', 'table_format')
+        self.syntax_style = c.get('main', 'syntax_style')
 
         self.logger = logging.getLogger(__name__)
         self.initialize_logging()
 
         self.query_history = []
 
+        # Initialize completer
+        smart_completion = c.getboolean('main', 'smart_completion')
+        completer = PGCompleter(smart_completion)
+        completer.extend_special_commands(CASE_SENSITIVE_COMMANDS.keys())
+        completer.extend_special_commands(NON_CASE_SENSITIVE_COMMANDS.keys())
+        self.completer = completer
+
     def initialize_logging(self):
 
         log_file = self.config.get('main', 'log_file')
@@ -87,7 +105,7 @@ class PGCli(object):
         root_logger.setLevel(level_map[log_level.upper()])
 
         root_logger.debug('Initializing pgcli logging.')
-        root_logger.debug('Log file "%s".' % log_file)
+        root_logger.debug('Log file %r.', log_file)
 
     def connect_uri(self, uri):
         uri = urlparse(uri)
@@ -126,7 +144,8 @@ class PGCli(object):
             try:
                 pgexecute = PGExecute(database, user, passwd, host, port)
             except OperationalError as e:
-                if 'no password supplied' in e.args[0] and auto_passwd_prompt:
+                if ('no password supplied' in utf8tounicode(e.args[0]) and
+                        auto_passwd_prompt):
                     passwd = click.prompt('Password', hide_input=True,
                                           show_default=False, type=str)
                     pgexecute = PGExecute(database, user, passwd, host, port)
@@ -135,6 +154,7 @@ class PGCli(object):
 
         except Exception as e:  # Connecting to a database could fail.
             self.logger.debug('Database connection failed: %r.', e)
+            self.logger.error("traceback: %r", traceback.format_exc())
             click.secho(str(e), err=True, fg='red')
             exit(1)
 
@@ -146,18 +166,23 @@ class PGCli(object):
         logger = self.logger
         original_less_opts = self.adjust_less_opts()
 
+        completer = self.completer
+        self.refresh_completions()
+        key_binding_manager = pgcli_bindings(self.vi_mode)
+        print('Version:', __version__)
+        print('Chat: https://gitter.im/amjith/pgcli')
+        print('Mail: https://groups.google.com/forum/#!forum/pgcli')
+        print('Home: http://pgcli.com')
+
         layout = Layout(before_input=DefaultPrompt(prompt),
             menus=[CompletionsMenu(max_height=10)],
-            lexer=SqlLexer, bottom_toolbars=[PGToolbar()])
-
-        completer = PGCompleter(self.smart_completion)
-        completer.extend_special_commands(CASE_SENSITIVE_COMMANDS.keys())
-        completer.extend_special_commands(NON_CASE_SENSITIVE_COMMANDS.keys())
-        refresh_completions(pgexecute, completer)
+            lexer=PostgresLexer,
+            bottom_toolbars=[PGToolbar(key_binding_manager)])
         buf = PGBuffer(always_multiline=self.multi_line, completer=completer,
                 history=FileHistory(os.path.expanduser('~/.pgcli-history')))
-        cli = CommandLineInterface(style=PGStyle, layout=layout, buffer=buf,
-                key_bindings_registry=pgcli_bindings())
+        cli = CommandLineInterface(style=style_factory(self.syntax_style),
+                layout=layout, buffer=buf,
+                key_bindings_registry=key_binding_manager.registry)
 
         try:
             while True:
@@ -178,16 +203,37 @@ class PGCli(object):
 
                 try:
                     logger.debug('sql: %r', document.text)
+                    successful = False
+                    # Initialized to [] because res might never get initialized
+                    # if an exception occurs in pgexecute.run(). Which causes
+                    # finally clause to fail.
+                    res = []
+                    start = time()
+                    # Run the query.
                     res = pgexecute.run(document.text)
+                    duration = time() - start
                     successful = True
                     output = []
-                    for rows, headers, status in res:
+                    total = 0
+                    for title, cur, headers, status in res:
                         logger.debug("headers: %r", headers)
-                        logger.debug("rows: %r", rows)
+                        logger.debug("rows: %r", cur)
                         logger.debug("status: %r", status)
-                        output.extend(format_output(rows, headers, status))
+                        start = time()
+                        threshold = 1000
+                        if (is_select(status) and
+                                cur and cur.rowcount > threshold):
+                            click.secho('The result set has more than %s rows.'
+                                    % threshold, fg='red')
+                            if not click.confirm('Do you want to continue?'):
+                                click.secho("Aborted!", err=True, fg='red')
+                                break
+                        output.extend(format_output(title, cur, headers,
+                            status, self.table_format))
+                        end = time()
+                        total += end - start
                         mutating = mutating or is_mutating(status)
-                    click.echo_via_pager('\n'.join(output))
+
                 except KeyboardInterrupt:
                     # Restart connection to the database
                     pgexecute.connect()
@@ -197,13 +243,26 @@ class PGCli(object):
                     logger.error("sql: %r, error: %r", document.text, e)
                     logger.error("traceback: %r", traceback.format_exc())
                     click.secho(str(e), err=True, fg='red')
-                    successful = False
+                else:
+                    click.echo_via_pager('\n'.join(output))
+                    if pgspecial.TIMING_ENABLED:
+                        print('Command Time:', duration)
+                        print('Format Time:', total)
+                finally:
+                    for cur, _, _ in res:
+                        if hasattr(cur, 'close'):
+                            cur.close()
 
                 # Refresh the table names and column names if necessary.
                 if need_completion_refresh(document.text):
                     prompt = '%s> ' % pgexecute.dbname
-                    completer.reset_completions()
-                    refresh_completions(pgexecute, completer)
+                    self.refresh_completions()
+
+                # Refresh search_path to set default schema.
+                if need_search_path_refresh(document.text):
+                    logger.debug('Refreshing search path')
+                    completer.set_search_path(pgexecute.search_path())
+                    logger.debug('Search path: %r', completer.search_path)
 
                 query = Query(document.text, successful, mutating)
                 self.query_history.append(query)
@@ -217,16 +276,32 @@ class PGCli(object):
     def adjust_less_opts(self):
         less_opts = os.environ.get('LESS', '')
         self.logger.debug('Original value for LESS env var: %r', less_opts)
-        if not less_opts:
-            os.environ['LESS'] = '-RXF'
+        os.environ['LESS'] = '-RXF'
 
-        if 'X' not in less_opts:
-            os.environ['LESS'] += 'X'
-        if 'F' not in less_opts:
-            os.environ['LESS'] += 'F'
+        #if 'X' not in less_opts:
+            #os.environ['LESS'] += 'X'
+        #if 'F' not in less_opts:
+            #os.environ['LESS'] += 'F'
 
         return less_opts
 
+    def refresh_completions(self):
+        completer = self.completer
+        completer.reset_completions()
+
+        pgexecute = self.pgexecute
+
+        completer.set_search_path(pgexecute.search_path())
+        completer.extend_schemata(pgexecute.schemata())
+        completer.extend_tables(pgexecute.tables())
+        completer.extend_columns(pgexecute.columns())
+        completer.extend_functions(pgexecute.functions())
+        completer.extend_database_names(pgexecute.databases())
+
+    def get_completions(self, text, cursor_positition):
+        return self.completer.get_completions(
+            Document(text=text, cursor_position=cursor_positition), None)
+
 @click.command()
 # Default host is '' so psycopg2 can default to either localhost or unix socket
 @click.option('-h', '--host', default='', envvar='PGHOST',
@@ -239,12 +314,18 @@ class PGCli(object):
         help='Force password prompt.')
 @click.option('-w', '--no-password', 'never_prompt', is_flag=True,
         default=False, help='Never prompt for password.')
+@click.option('-v', '--version', is_flag=True, help='Version of pgcli.')
 @click.option('-d', '--dbname', default='', envvar='PGDATABASE',
         help='database name to connect to.')
 @click.argument('database', default=lambda: None, envvar='PGDATABASE', nargs=1)
 @click.argument('username', default=lambda: None, envvar='PGUSER', nargs=1)
 def cli(database, user, host, port, prompt_passwd, never_prompt, dbname,
-        username):
+        username, version):
+
+    if version:
+        print('Version:', __version__)
+        sys.exit(0)
+
     pgcli = PGCli(prompt_passwd, never_prompt)
 
     # Choose which ever one has a valid value.
@@ -264,44 +345,56 @@ def cli(database, user, host, port, prompt_passwd, never_prompt, dbname,
 
     pgcli.run_cli()
 
-def format_output(rows, headers, status):
+def format_output(title, cur, headers, status, table_format):
     output = []
-    if rows:
+    if title:  # Only print the title if it's not None.
+        output.append(title)
+    if cur:
+        headers = [utf8tounicode(x) for x in headers]
         if is_expanded_output():
-            output.append(expanded_table(rows, headers))
+            output.append(expanded_table(cur, headers))
         else:
-            output.append(tabulate(rows, headers, tablefmt='psql'))
+            output.append(tabulate(cur, headers, tablefmt=table_format,
+                missingval='<null>'))
     if status:  # Only print the status if it's not None.
         output.append(status)
     return output
 
-def need_completion_refresh(sql):
-    try:
-        first_token = sql.split()[0]
-        return first_token.lower() in ('alter', 'create', 'use', '\c', 'drop')
-    except Exception:
-        return False
+def need_completion_refresh(queries):
+    """Determines if the completion needs a refresh by checking if the sql
+    statement is an alter, create, drop or change db."""
+    for query in sqlparse.split(queries):
+        try:
+            first_token = query.split()[0]
+            return first_token.lower() in ('alter', 'create', 'use', '\\c',
+                    '\\connect', 'drop')
+        except Exception:
+            return False
+
+def need_search_path_refresh(sql):
+    """Determines if the search_path should be refreshed by checking if the
+    sql has 'set search_path'."""
+    return 'set search_path' in sql.lower()
 
 def is_mutating(status):
+    """Determines if the statement is mutating based on the status."""
     if not status:
         return False
-    
-    mutating = ['insert', 'update', 'delete', 'alter', 'create', 'drop']
+
+    mutating = set(['insert', 'update', 'delete', 'alter', 'create', 'drop'])
     return status.split(None, 1)[0].lower() in mutating
 
+def is_select(status):
+    """Returns true if the first word in status is 'select'."""
+    if not status:
+        return False
+    return status.split(None, 1)[0].lower() == 'select'
+
 def quit_command(sql):
     return (sql.strip().lower() == 'exit'
             or sql.strip().lower() == 'quit'
             or sql.strip() == '\q'
             or sql.strip() == ':q')
 
-def refresh_completions(pgexecute, completer):
-    tables, columns = pgexecute.tables()
-    completer.extend_table_names(tables)
-    for table in tables:
-        table = table[1:-1] if table[0] == '"' and table[-1] == '"' else table
-        completer.extend_column_names(table, columns[table])
-    completer.extend_database_names(pgexecute.databases())
-
 if __name__ == "__main__":
     cli()
