@@ -11,18 +11,12 @@ TableInfo = namedtuple("TableInfo", ['checks', 'relkind', 'hasindex',
 
 log = logging.getLogger(__name__)
 
-class MockLogging(object):
-    def debug(self, string):
-        print ("***** Query ******")
-        print (string)
-        print ("******************")
-        print ()
-
-#log = MockLogging()
 use_expanded_output = False
 def is_expanded_output():
     return use_expanded_output
 
+TIMING_ENABLED = False
+
 def parse_special_command(sql):
     command, _, arg = sql.partition(' ')
     verbose = '+' in command
@@ -54,7 +48,145 @@ def list_schemas(cur, pattern, verbose):
     cur.execute(sql)
     if cur.description:
         headers = [x[0] for x in cur.description]
-        return [(cur.fetchall(), headers, cur.statusmessage)]
+        return [(None, cur, headers, cur.statusmessage)]
+
+def list_objects(cur, pattern, verbose, relkinds):
+    """
+        Returns (rows, header, status)
+
+        This method is used by list_tables, list_views, and list_indexes
+
+        relkinds is a list of strings to filter pg_class.relkind
+
+    """
+    schema_pattern, table_pattern = sql_name_pattern(pattern)
+
+    if verbose:
+        verbose_columns = '''
+            ,pg_catalog.pg_size_pretty(pg_catalog.pg_table_size(c.oid)) as "Size",
+            pg_catalog.obj_description(c.oid, 'pg_class') as "Description" '''
+    else:
+        verbose_columns = ''
+
+    sql = '''SELECT n.nspname as "Schema",
+                    c.relname as "Name",
+                    CASE c.relkind
+                      WHEN 'r' THEN 'table' WHEN 'v' THEN 'view'
+                      WHEN 'm' THEN 'materialized view' WHEN 'i' THEN 'index'
+                      WHEN 'S' THEN 'sequence' WHEN 's' THEN 'special'
+                      WHEN 'f' THEN 'foreign table' END
+                    as "Type",
+                    pg_catalog.pg_get_userbyid(c.relowner) as "Owner"
+          ''' + verbose_columns + '''
+            FROM    pg_catalog.pg_class c
+                    LEFT JOIN pg_catalog.pg_namespace n
+                      ON n.oid = c.relnamespace
+            WHERE   c.relkind = ANY(%s) '''
+
+    params = [relkinds]
+
+    if schema_pattern:
+        sql += ' AND n.nspname ~ %s'
+        params.append(schema_pattern)
+    else:
+        sql += '''
+            AND n.nspname <> 'pg_catalog'
+            AND n.nspname <> 'information_schema'
+            AND n.nspname !~ '^pg_toast'
+            AND pg_catalog.pg_table_is_visible(c.oid) '''
+
+    if table_pattern:
+        sql += ' AND c.relname ~ %s'
+        params.append(table_pattern)
+
+    sql = cur.mogrify(sql + ' ORDER BY 1, 2', params)
+
+    log.debug(sql)
+    cur.execute(sql)
+
+    if cur.description:
+        headers = [x[0] for x in cur.description]
+        return [(None, cur, headers, cur.statusmessage)]
+
+
+def list_tables(cur, pattern, verbose):
+    return list_objects(cur, pattern, verbose, ['r', ''])
+
+
+def list_views(cur, pattern, verbose):
+    return list_objects(cur, pattern, verbose, ['v', 's', ''])
+
+def list_sequences(cur, pattern, verbose):
+    return list_objects(cur, pattern, verbose, ['S', 's', ''])
+
+def list_indexes(cur, pattern, verbose):
+    return list_objects(cur, pattern, verbose, ['i', 's', ''])
+
+
+def list_functions(cur, pattern, verbose):
+
+    if verbose:
+        verbose_columns = '''
+            ,CASE
+                 WHEN p.provolatile = 'i' THEN 'immutable'
+                 WHEN p.provolatile = 's' THEN 'stable'
+                 WHEN p.provolatile = 'v' THEN 'volatile'
+            END as "Volatility",
+            pg_catalog.pg_get_userbyid(p.proowner) as "Owner",
+          l.lanname as "Language",
+          p.prosrc as "Source code",
+          pg_catalog.obj_description(p.oid, 'pg_proc') as "Description" '''
+
+        verbose_table = ''' LEFT JOIN pg_catalog.pg_language l
+                                ON l.oid = p.prolang'''
+    else:
+        verbose_columns = verbose_table = ''
+
+    sql = '''
+        SELECT  n.nspname as "Schema",
+                p.proname as "Name",
+                pg_catalog.pg_get_function_result(p.oid)
+                  as "Result data type",
+                pg_catalog.pg_get_function_arguments(p.oid)
+                  as "Argument data types",
+                 CASE
+                    WHEN p.proisagg THEN 'agg'
+                    WHEN p.proiswindow THEN 'window'
+                    WHEN p.prorettype = 'pg_catalog.trigger'::pg_catalog.regtype
+                        THEN 'trigger'
+                    ELSE 'normal'
+                END as "Type" ''' + verbose_columns + '''
+        FROM    pg_catalog.pg_proc p
+                LEFT JOIN pg_catalog.pg_namespace n ON n.oid = p.pronamespace
+                        ''' + verbose_table + '''
+        WHERE  '''
+
+    schema_pattern, func_pattern = sql_name_pattern(pattern)
+    params = []
+
+    if schema_pattern:
+        sql += ' n.nspname ~ %s '
+        params.append(schema_pattern)
+    else:
+        sql += ' pg_catalog.pg_function_is_visible(p.oid) '
+
+    if func_pattern:
+        sql += ' AND p.proname ~ %s '
+        params.append(func_pattern)
+
+    if not (schema_pattern or func_pattern):
+        sql += ''' AND n.nspname <> 'pg_catalog'
+                   AND n.nspname <> 'information_schema' '''
+
+    sql = cur.mogrify(sql + ' ORDER BY 1, 2, 4', params)
+
+    log.debug(sql)
+    cur.execute(sql)
+
+    if cur.description:
+        headers = [x[0] for x in cur.description]
+        return [(None, cur, headers, cur.statusmessage)]
+
 
 def describe_table_details(cur, pattern, verbose):
     """
@@ -86,7 +218,7 @@ def describe_table_details(cur, pattern, verbose):
         cur.execute(sql)
         if cur.description:
             headers = [x[0] for x in cur.description]
-            return [(cur.fetchall(), headers, cur.statusmessage)]
+            return [(None, cur, headers, cur.statusmessage)]
 
     # This is a \d <tablename> command. A royal pain in the ass.
     schema, relname = sql_name_pattern(pattern)
@@ -116,7 +248,7 @@ def describe_table_details(cur, pattern, verbose):
     log.debug(sql)
     cur.execute(sql)
     if not (cur.rowcount > 0):
-        return [(None, None, 'Did not find any relation named %s.' % pattern)]
+        return [(None, None, None, 'Did not find any relation named %s.' % pattern)]
 
     results = []
     for oid, nspname, relname in cur.fetchall():
@@ -150,7 +282,7 @@ def describe_one_table_details(cur, schema_name, relation_name, oid, verbose):
     if (cur.rowcount > 0):
         tableinfo = TableInfo._make(cur.fetchone())
     else:
-        return (None, None, 'Did not find any relation with OID %s.' % oid)
+        return (None, None, None, 'Did not find any relation with OID %s.' % oid)
 
     # If it's a seq, fetch it's value and store it for later.
     if tableinfo.relkind == 'S':
@@ -159,7 +291,7 @@ def describe_one_table_details(cur, schema_name, relation_name, oid, verbose):
         log.debug(sql)
         cur.execute(sql)
         if not (cur.rowcount > 0):
-            return (None, None, 'Something went wrong.')
+            return (None, None, None, 'Something went wrong.')
 
         seq_values = cur.fetchone()
 
@@ -707,7 +839,7 @@ def describe_one_table_details(cur, schema_name, relation_name, oid, verbose):
     if (verbose and tableinfo.reloptions):
         status.append("Options: %s\n" % tableinfo.reloptions)
 
-    return (cells, headers, "".join(status))
+    return (None, cells, headers, "".join(status))
 
 def sql_name_pattern(pattern):
     """
@@ -766,7 +898,7 @@ def show_help(cur, arg, verbose):  # All the parameters are ignored.
     for command, value in sorted(CASE_SENSITIVE_COMMANDS.items()):
         if value[1]:
             result.append(value[1])
-    return [(result, headers, None)]
+    return [(None, result, headers, None)]
 
 def change_db(cur, arg, verbose):
     raise NotImplementedError
@@ -775,8 +907,15 @@ def expanded_output(cur, arg, verbose):
     global use_expanded_output
     use_expanded_output = not use_expanded_output
     message = u"Expanded display is "
-    message += u"on" if use_expanded_output else u"off"
-    return [(None, None, message + u".")]
+    message += u"on." if use_expanded_output else u"off."
+    return [(None, None, None, message)]
+
+def toggle_timing(cur, arg, verbose):
+    global TIMING_ENABLED
+    TIMING_ENABLED = not TIMING_ENABLED
+    message = "Timing is "
+    message += "on." if TIMING_ENABLED else "off."
+    return [(None, None, None, message)]
 
 CASE_SENSITIVE_COMMANDS = {
             '\?': (show_help, ['\?', 'Help on pgcli commands.']),
@@ -785,40 +924,12 @@ CASE_SENSITIVE_COMMANDS = {
             '\d': (describe_table_details, ['\d [pattern]', 'list or describe tables, views and sequences.']),
             '\dn': (list_schemas, ['\dn[+] [pattern]', 'list schemas']),
             '\\x': (expanded_output, ['\\x', 'Toggle expanded output.']),
-            '\dt': ('''SELECT n.nspname as "Schema", c.relname as "Name", CASE
-            c.relkind WHEN 'r' THEN 'table' WHEN 'v' THEN 'view' WHEN 'm' THEN
-            'materialized view' WHEN 'i' THEN 'index' WHEN 'S' THEN 'sequence'
-            WHEN 's' THEN 'special' WHEN 'f' THEN 'foreign table' END as
-            "Type", pg_catalog.pg_get_userbyid(c.relowner) as "Owner" FROM
-            pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_namespace n ON n.oid
-            = c.relnamespace WHERE c.relkind IN ('r','') AND n.nspname <>
-            'pg_catalog' AND n.nspname <> 'information_schema' AND n.nspname !~
-            '^pg_toast' AND pg_catalog.pg_table_is_visible(c.oid) ORDER BY
-            1,2;''', ['\dt', 'list tables.']),
-            '\di': ('''SELECT n.nspname as "Schema", c.relname as "Name", CASE
-            c.relkind WHEN 'r' THEN 'table' WHEN 'v' THEN 'view' WHEN 'm' THEN
-            'materialized view' WHEN 'i' THEN 'index' WHEN 'S' THEN 'sequence'
-            WHEN 's' THEN 'special' WHEN 'f' THEN 'foreign table' END as
-            "Type", pg_catalog.pg_get_userbyid(c.relowner) as "Owner",
-            c2.relname as "Table" FROM pg_catalog.pg_class c LEFT JOIN
-            pg_catalog.pg_namespace n ON n.oid = c.relnamespace LEFT JOIN
-            pg_catalog.pg_index i ON i.indexrelid = c.oid LEFT JOIN
-            pg_catalog.pg_class c2 ON i.indrelid = c2.oid WHERE c.relkind
-            IN ('i','') AND n.nspname <> 'pg_catalog' AND
-            n.nspname <> 'information_schema' AND n.nspname !~ '^pg_toast'
-            AND pg_catalog.pg_table_is_visible(c.oid)
-            ORDER BY 1,2;''', ['\di', 'list indexes.']),
-            '\dv': ('''SELECT n.nspname as "Schema", c.relname as "Name", CASE
-            c.relkind WHEN 'r' THEN 'table' WHEN 'v' THEN 'view' WHEN 'm' THEN
-            'materialized view' WHEN 'i' THEN 'index' WHEN 'S' THEN 'sequence'
-            WHEN 's' THEN 'special' WHEN 'f' THEN 'foreign table' END as
-            "Type", pg_catalog.pg_get_userbyid(c.relowner) as "Owner" FROM
-            pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_namespace n ON
-            n.oid = c.relnamespace WHERE c.relkind IN ('v','') AND
-            n.nspname <> 'pg_catalog' AND n.nspname <> 'information_schema'
-            AND n.nspname !~ '^pg_toast' AND
-            pg_catalog.pg_table_is_visible(c.oid) ORDER BY 1,2;''',
-            ['\dv', 'list views.']),
+            '\\timing': (toggle_timing, ['\\timing', 'Toggle timing of commands.']),
+            '\\dt': (list_tables, ['\\dt[+] [pattern]', 'list tables.']),
+            '\\di': (list_indexes, ['\\di[+] [pattern]', 'list indexes.']),
+            '\\dv': (list_views, ['\\dv[+] [pattern]', 'list views.']),
+            '\\ds': (list_sequences, ['\\ds[+] [pattern]', 'list sequences.']),
+            '\\df': (list_functions, ['\\df[+] [pattern]', 'list functions.'])
             }
 
 NON_CASE_SENSITIVE_COMMANDS = {
@@ -848,9 +959,9 @@ def execute(cur, sql):
         cur.execute(command_executor)
         if cur.description:
             headers = [x[0] for x in cur.description]
-            return [(cur.fetchall(), headers, cur.statusmessage)]
+            return [(None, cur, headers, cur.statusmessage)]
         else:
-            return [(None, None, cur.statusmessage)]
+            return [(None, None, None, cur.statusmessage)]
 
 if __name__ == '__main__':
     import psycopg2
