@@ -5,6 +5,7 @@
 from __future__ import print_function
 from __future__ import unicode_literals
 from collections import namedtuple
+from decimal import Decimal
 from platform import python_version_tuple
 import re
 
@@ -14,6 +15,7 @@ if python_version_tuple()[0] < "3":
     from functools import partial
     _none_type = type(None)
     _int_type = int
+    _long_type = long
     _float_type = float
     _text_type = unicode
     _binary_type = str
@@ -26,6 +28,7 @@ else:
     from functools import reduce, partial
     _none_type = type(None)
     _int_type = int
+    _long_type = int
     _float_type = float
     _text_type = str
     _binary_type = bytes
@@ -302,7 +305,7 @@ def _isint(string):
     >>> _isint("123.45")
     False
     """
-    return type(string) is int or \
+    return type(string) is _int_type or type(string) is _long_type or \
            (isinstance(string, _binary_type) or isinstance(string, _text_type)) and \
            _isconvertible(int, string)
 
@@ -329,6 +332,8 @@ def _type(string, has_invisible=True):
 
     if string is None:
         return _none_type
+    if isinstance(string, (bool, Decimal,)):
+        return _text_type
     elif hasattr(string, "isoformat"):  # datetime.datetime, date, and time
         return _text_type
     elif _isint(string):
